from typing import Callable

from jax import grad, jit


# TODO: Imperfect test. Need to be able to call grad_func on some default values
# (e.g., lambda x: onp.random.rand(x) won't error here, but should)
def differentiable(func: Callable, *args, **kwargs) -> bool:
    if not callable(func):
        return False

    try:
        grad_func = grad(func)
        result = grad_func(*args)
        jit(grad_func)
        result
    except Exception as e:
        print(e)
        # TODO: What is actual jax exception?
        return False

    return True



from numbers import Real

import jax.numpy as np

from proust.utils.losses.core import Loss


class CrossEntropy(Loss):
    def __init__(self, eps: Real = 1e-9):
        self._eps = eps

    def compute(self, y_pred: np.ndarray, y_true: np.ndarray):
        """
        Description: cross entropy loss, y_pred is equivalent to logits and
        y_true to labels
        Args:
            y_pred (np.ndarray): value predicted by method
            y_true (np.ndarray): ground truth value
        Returns:
            error (Real):
        """
        return -np.sum(y_true * np.log(y_pred + self._eps)) / y_pred.shape[0]



"""
AR: a linear auto-regressive learner

Assumptions
- Time series values have the same dimension as the predicted value or is scalar
- In the scalar scenario, all features in the input_dim x window_size history
  are flattened to predict the single scalar value (input_dim x window_size features)
- In the vector scenario, each input_dim-sized vector in the window_size history
  is treated as its own regressor (window_size features)
"""

from numbers import Real
from typing import Generator, Tuple, Union

import jax
import jax.experimental.stax as stax
import jax.numpy as np
import jax.ops
import numpy as onp

from proust.learners.core import (
    BaseLearner,
    DifferentiableMixin,
    OfflineMixin,
    RegressorMixin,
    TimeSeries,
    TimeSeriesMixin,
)
from proust.utils.losses import MeanSquareError
from proust.utils.losses.core import Loss
from proust.utils.optimizers import OGD
from proust.utils.optimizers.core import Optimizer
from proust.utils.random import generate_key
from proust.utils.statistics import OnlineStatistics

# TODO: should we normalize online data as it's coming in if we don't have any
# pre-training?
# TODO: add constrained flag


# TODO: Type hint data
def _ar_update_history(history: np.ndarray, data) -> np.ndarray:
    history = np.asarray(history)
    if history.shape[1] != len(data):
        raise ValueError(
            "Attempting to update history with time series value of differing shape."
        )
    _updated_history = np.roll(history, len(data))
    _updated_history = jax.ops.index_update(_updated_history, 0, data)
    return _updated_history


def _ar_predict(
    params: dict, history: np.ndarray, fit_intercept: bool = True
) -> np.ndarray:
    """
    Description: Takes in a history of n-d vectors and predicts an n-d vector.
    Treats each n-d vector in the history as a multi-dimensional regressor.

    Args:

        params (dict): Currently implemented as a dictionary with key 'phi'
        that holds an (window_size + 1)-d vector of parameters for a history of
        window_size n-d vectors
        history (np.ndarray): A history of n-d vectors; two-dimensional array
        where the first dimension is time and the second dimension (if any) is
        the time series value, which may be multidimensional
        fit_intercept (bool): whether or not to fit an intercept

    Raises:
        ValueError: [description]
        ValueError: [description]

    Returns:
        np.ndarray: An n-d prediction
    """

    # - Adds a vector of ones as an additiona n-d vector to history
    # - Matrix multiplies with (n + 1) parameters to return prediction
    history = np.asarray(history)

    if fit_intercept:
        return np.dot(
            np.vstack((np.ones((1, history.shape[1])), history)).T, params["phi"]
        )
    else:
        return np.dot(history.T, params["phi"])


def _ar_predict_single(
    params: dict, history: np.ndarray, fit_intercept: bool = True
) -> Real:
    """
    Description: Takes in a history of n-d vectors and predicts a scalar value.
    Flattens the history to use as a (window_size * input_dim + 1)-dimensional
    regression.

    Args:
        params (dict): Currently implemented as a dictionary with key 'phi'
        that holds a (window_size * input_dim + 1)-d vector of parameters.
        history (np.ndarray): A history of n-d vectors that is window_size by
        input_dim
        fit_intercept (bool): whether or not to fit an intercept

    Returns:
        np.ndarray: [description]
    """
    if fit_intercept:
        return np.dot(np.concatenate((np.array([1]), history.ravel())), params["phi"])
    else:
        return np.dot(history.ravel(), params["phi"])


def _ar_form_linear_constraints(
    input_dim: int, output_dim: int, window_size: int, fit_intercept: bool = True
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Description: Sets the linear constraint matrix and vector R and r,
    respectively for constraining the `_params` vector. Specifically, each row
    in R is a linearly independent constraint where we force each dimension in
    the time series value to share a coefficient across the window.

    References: https://www.le.ac.uk/users/dsgp1/COURSES/TOPICS/restrict.pdf

    Args:
        input_dim (int): number of dimensions in the time series value
        output_dim (int): number of dimensions in the output
        window_size (int): window size for the auto reggressor
        fit_intercept (bool, optional): whether or not to fit an intercept.
        Defaults to True.

    Returns:
        Tuple[np.ndarray, np.ndarray]: RB = r
    """
    if fit_intercept:
        window_size += 1

    num_constraints = window_size * (input_dim - 1)

    R = onp.zeros((num_constraints, input_dim * window_size))
    r = onp.zeros((num_constraints, output_dim))

    # To share a parameter p across features x1 and x2, create a new row in R
    # with index 1 set to 1, index 2 set to -1, the rest set to 0 and the
    # corresponding row in r set to 0 (i.e., x1 + x2 = 0)
    for i in range(window_size):
        for j in range(input_dim - 1):

            # Set the constraint row
            row = i * (input_dim - 1) + j

            # Column indices for the first and second feature to tie. Note that
            # within the inner for loop, we tie multiple features to the same
            # parameter
            col1 = i * input_dim + j
            col2 = i * input_dim + j + 1

            # Update R appropriately
            R[row, col1] = 1
            R[row, col2] = -1

    return R, r


def _ar_fit_normalize_prod(
    # inv: np.ndarray, mu: np.ndarray, sigma: np.ndarray, s: np.ndarray, n: int
    AB: np.ndarray,
    A_stats: OnlineStatistics,
    B_stats: OnlineStatistics,
) -> np.ndarray:
    """
    Normalize the product of matrices A and B if we only have running mean, std,
    sum

    Args:
        AB (np.ndarray): The product of matrix A and B
        A_stats (OnlineStatistics): running mean, std, sum for matrix A
        B_stats (OnlineStatistics): running mean, std, sum for matrix B

    Raises:
        ValueError: [description]

    Returns:
        np.ndarray: Normalized AB
    """

    if A_stats.observations() != B_stats.observations():
        raise ValueError("A and B should have the same number of observations")

    n = A_stats.observations()

    result = (
        AB
        + n * np.outer(A_stats.mean(), B_stats.mean())
        - np.outer(A_stats.mean(), B_stats.sum())
        - np.outer(A_stats.sum(), B_stats.mean())
    ) / np.outer(A_stats.std(), B_stats.std())

    # Replace nans with 0s because std of some column was 0
    return np.nan_to_num(result)


def _ar_fit_unconstrained(inv: np.ndarray, Xy: np.ndarray) -> np.ndarray:
    print(inv.shape, Xy.shape)
    return inv @ Xy


def _ar_fit_constrained(beta, inv, R, r):
    print(beta.shape, inv.shape, R.shape, r.shape)
    return beta - inv @ (R.T @ onp.linalg.inv(R @ (inv @ R.T))) @ (R @ beta - r)


class AR(
    DifferentiableMixin, OfflineMixin, RegressorMixin, TimeSeriesMixin, BaseLearner,
):
    def __init__(
        self,
        input_dim: int = 1,
        output_dim: int = 1,
        window_size: int = 3,
        optimizer: Optimizer = OGD,
        loss: Loss = None,
        fit_intercept: bool = True,
        normalize: bool = False,
        constrain: bool = False,
    ):
        super().__init__()

        self._input_dim = input_dim
        self._output_dim = output_dim
        self._window_size = window_size
        self._optimizer = optimizer
        self._loss = MeanSquareError() if loss is None else loss
        self._fit_intercept = fit_intercept
        self._normalize = normalize
        self._constrain = constrain

        self._num_features = self._input_dim * self._window_size
        self._num_features_with_intercepts = self._num_features

        # If we want to fit an intercept, need to add features
        if self._fit_intercept:
            # If we're constraining so each input_dim vector in the window is
            # its own regressor, add input_dim intercepts
            if self._constrain:
                self._num_features_with_intercepts += self._input_dim

            # If unconstrained, add 1
            else:
                self._num_features_with_intercepts += 1

        self._param_dim = [self._window_size, 1]
        if self._fit_intercept:
            self._param_dim[0] += 1

        # TODO: The great (R,), (R,1) debate rages on! Here, we've chosen
        #  (R,) in the case of scalar time series values
        self._history_shape = (self._window_size, self._input_dim)
        self._history = np.zeros(self._history_shape)

        glorot_init = stax.glorot()
        self._params = {"phi": glorot_init(generate_key(), self._param_dim)}

        self._update_history = jax.jit(
            lambda history, data: _ar_update_history(history, data)
        )

        self._predict = jax.jit(
            lambda params, data: _ar_predict(params, data, self._fit_intercept)
        )

        self._objective = jax.jit(
            lambda params, x, y: self._loss.compute(self._predict(params, x), y)
        )

        self._grad = jax.jit(jax.grad(self._objective))

        self._X_stats = OnlineStatistics(self._window_size * self._input_dim)
        self._y_stats = OnlineStatistics(self._output_dim)

    def predict(self, data: TimeSeries) -> TimeSeries:
        if self._fit_intercept and self._normalize:
            data = self._X_stats.zscore(data)
        # NOTE: We call self._update_history twice each iteration; once in
        # predict, once in update
        return self._predict(self._params, self._update_history(self._history, data))

    def update(self, data: TimeSeries, y) -> "AR":
        new_data = data
        if self._fit_intercept and self._normalize:
            self._X_stats.update(data)
            new_data = self._X_stats.zscore(data)

        # NOTE: update both state and params here. Must be run using the same
        # data as in predict (preferably right after running predict)
        self._history = self._update_history(self._history, new_data)
        self._params = self._optimizer.update(
            self._params, self._grad(self._params, data, y)
        )
        return self

    # `fit` helper function split out for testing
    def _fit_accumulate(
        self,
        generator: Generator[Tuple[np.ndarray, np.ndarray], None, None],
        alpha: Union[Real, np.ndarray] = 1,
    ) -> (np.ndarray, np.ndarray):

        XTX = onp.zeros((self._num_features, self._num_features))
        Xy = onp.zeros((self._num_features, self._output_dim))

        for X, y in generator:
            # We don't use num_features here because this is data unmodified by
            # intercepts
            if X.shape[1] != self._num_features:
                raise ValueError(
                    "Generator provided X {} that has shape inconsistent with"
                    "`input_dim` {} and `window_size` {}".format(
                        X.shape, self._input_dim, self._window_size
                    )
                )

            if len(y.shape) > 1 and y.shape[1] != self._output_dim:
                raise ValueError(
                    "Generator provided y {} inconsistent with `output_dim` {}".format(
                        y.shape, self._output_dim
                    )
                )

            # Force vectors into matrices
            if len(X.shape) == 1:
                X = X[:, np.newaxis]
            if len(y.shape) == 1:
                y = y[:, np.newaxis]

            XTX += X.T @ X
            Xy += X.T @ y

            for (X_, y_) in zip(X, y):
                self._X_stats.update(X_)
                self._y_stats.update(y_)

        if self._fit_intercept:
            if self._normalize:

                XTX = _ar_fit_normalize_prod(XTX, self._X_stats, self._X_stats)
                Xy = _ar_fit_normalize_prod(Xy, self._X_stats, self._y_stats)

            num_rc = self._input_dim if self._constrain else 1

            # pylint: disable=unsubscriptable-object
            XTX = onp.concatenate((np.zeros((XTX.shape[0], num_rc)), XTX), axis=1)
            XTX = onp.concatenate((np.zeros((num_rc, XTX.shape[1])), XTX), axis=0)
            Xy = onp.concatenate((np.zeros((num_rc, Xy.shape[1])), Xy), axis=0)

            if not self._normalize:
                S = onp.tile(self._X_stats.sum(), (num_rc, 1))
                XTX[:num_rc, :num_rc] = self._X_stats.observations()
                XTX[:num_rc, num_rc:] = S
                XTX[num_rc:, :num_rc] = S.T

                S = onp.tile(self._y_stats.sum(), (num_rc, 1))
                Xy[:num_rc, :] = S

        inv = onp.linalg.inv(XTX + alpha * np.eye(XTX.shape[1]))

        return inv, Xy

    def fit(
        self,
        generator: Generator[Tuple[np.ndarray, np.ndarray], None, None],
        alpha: Real = 1,
    ) -> "AR":
        """
        Description: Incrementally computes a ridge regression given data yielded
        by `generator`. NOTE: this will retrain the underlying predictor from scratch.

        References: https://www.le.ac.uk/users/dsgp1/COURSES/TOPICS/restrict.pdf

        Args:
            generator (Generator[(np.ndarray, np.ndarray), None, None]): yields (X,
            y) tuples representing inputs and outputs, respectively. X is a
            two-dimensional array where the first dimension is number of
            observations and the second dimension is the flattened features
            alpha (Union[Real, np.ndarray], optional): [TODO]. Defaults to 1.

        Returns:
            self
        """

        # Sum up data incrementally
        inv, Xy = self._fit_accumulate(generator, alpha=alpha)
        self._params["phi"] = _ar_fit_unconstrained(inv, Xy)

        if self._constrain:
            R, r = _ar_form_linear_constraints(
                self._input_dim,
                self._output_dim,
                self._window_size,
                fit_intercept=self._fit_intercept,
            )
            self._params["phi"] = _ar_fit_constrained(self._params["phi"], inv, R, r)
            self._params["phi"] = self._params["phi"].take(
                np.arange(0, len(self._params["phi"]), self._input_dim), axis=0
            )

        return self



import numpy as np
import pytest

from proust.learners._predict_constant import PredictConstant


@pytest.mark.parametrize(
    "c,x,y",
    [
        (0, 0, 0),
        (0, np.array([0]), np.zeros(1)),
        (0, np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]), np.zeros(10)),
        (0, np.random.rand(2, 2), np.zeros((2, 2))),
        (0, np.random.rand(2, 2), np.zeros((2, 2))),
        (1, 0, 1),
        (1, np.array([0]), np.ones(1)),
        (1, np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]), np.ones(10)),
        (1, np.random.rand(2, 2), np.ones((2, 2))),
        (1, np.random.rand(2, 2), np.ones((2, 2))),
    ],
)
def test_predict_constant(c, x, y):
    predict_zero = PredictConstant(constant=c)

    assert np.array_equal(predict_zero.predict(x), y)



# pylint: disable=unsubscriptable-object
import jax.numpy as np
import numpy as onp
import pytest

from proust.learners._ar import (
    AR,
    _ar_fit_constrained,
    _ar_fit_normalize_prod,
    _ar_fit_unconstrained,
    _ar_form_linear_constraints,
    _ar_update_history,
)
from proust.utils.statistics import OnlineStatistics


def ols(X, y, input_dim, alpha=1, fit_intercept=True, normalize=False, constrain=False):
    if fit_intercept:
        X = onp.concatenate(
            (np.ones((X.shape[0], input_dim if constrain else 1)), X), axis=1
        )

        if normalize:
            X = onp.nan_to_num((X - X.mean(axis=0)) / X.std(axis=0))
            y = (y - y.mean(axis=0)) / y.std(axis=0)
    return onp.linalg.inv(X.T @ X + alpha * np.eye(X.shape[1])) @ (X.T @ y)


@pytest.mark.parametrize(
    "history,data,expected",
    [
        (
            np.array([[1, 2, 3], [3, 2, 4], [3, 6, 2]]),
            [2, 4, 5],
            np.array([[2, 4, 5], [1, 2, 3], [3, 2, 4]]),
        ),
        (
            [
                [1, 2, 3, 4],
                [4, 2, 5, 2],
                [5, 3, 2, 1],
                [5, 2, 3, 1],
                [4, 2, 1, 2],
                [1, 2, 3, 4],
                [1, 2, 5, 4],
            ],
            [4, 5, 2, 1],
            [
                [4, 5, 2, 1],
                [1, 2, 3, 4],
                [4, 2, 5, 2],
                [5, 3, 2, 1],
                [5, 2, 3, 1],
                [4, 2, 1, 2],
                [1, 2, 3, 4],
            ],
        ),
    ],
)
def test_update_history(history, data, expected):
    updated = _ar_update_history(history, data)
    assert np.array_equal(expected, updated)


def test_update_history_exception():
    with pytest.raises(ValueError):
        _ar_update_history(onp.random.rand(10, 4), [1])


# @pytest.mark.parametrize(
#     "params,data,fit_intercept,expected",
#     [
#         (np.array([0, 1, 6]), np.array([[1, 2, 3], [4, 5, 6]]), True, [25, 32, 39]),
#         (
#             np.array([0, 1, 6, 4, 1]),
#             np.array([[1, 2, 3], [4, 5, 6], [4, 5, 6], [2, 4, 2]]),
#             True,
#             [43, 56, 65],
#         ),
#     ],
# )
# def test_predict(params, data, fit_intercept, expected):
#     assert _ar_predict({"phi": params}, data, fit_intercept) == expected


def generator(X, y, chunk_size=100):
    chunks = X.shape[0] // chunk_size
    next_index = 0
    for i in range(chunks):
        curr_index = i * chunk_size
        next_index = (i + 1) * chunk_size
        if len(y) == len(y.ravel()):
            yield X[curr_index:next_index, :], y[curr_index:next_index]
        else:
            yield X[curr_index:next_index, :], y[curr_index:next_index, :]
    if next_index < X.shape[0]:
        if len(y) == len(y.ravel()):
            yield X[next_index:, :], y[next_index:]
        else:
            yield X[next_index:, :], y[next_index:, :]


@pytest.mark.parametrize("n", [10, 1000])
@pytest.mark.parametrize("input_dim", [1, 10])
@pytest.mark.parametrize("output_dim", [1, 10])
@pytest.mark.parametrize("window_size", [1, 10])
@pytest.mark.parametrize("fit_intercept", [True, False])
@pytest.mark.parametrize("normalize", [True, False])
def test_fit_accumulate(
    n, input_dim, output_dim, window_size, fit_intercept, normalize
):
    # NOTE: we use random data because we want to test dimensions and
    # correctness vs a second implementation
    X = onp.random.rand(n, input_dim * window_size)
    y = onp.random.rand(n, output_dim)

    ar = AR(
        input_dim,
        output_dim,
        window_size,
        fit_intercept=fit_intercept,
        normalize=normalize,
    )
    inv, Xy = ar._fit_accumulate(generator(X, y), alpha=1)

    if fit_intercept:
        X = onp.concatenate((onp.ones((X.shape[0], 1)), X), axis=1)

        if normalize:
            X = onp.nan_to_num((X - X.mean(axis=0)) / X.std(axis=0))
            y = (y - y.mean(axis=0)) / y.std(axis=0)

    result = onp.linalg.inv(X.T @ X + np.eye(inv.shape[1]))
    onp.testing.assert_array_almost_equal(result, inv, decimal=2)
    onp.testing.assert_array_almost_equal(X.T @ y, Xy, decimal=2)


@pytest.mark.parametrize("n", [10, 1000])
@pytest.mark.parametrize("input_dim", [1, 10])
@pytest.mark.parametrize("output_dim", [1, 10])
@pytest.mark.parametrize("window_size", [1, 10])
@pytest.mark.parametrize("fit_intercept", [True, False])
@pytest.mark.parametrize("normalize", [True, False])
def test_fit_unconstrained(
    n, input_dim, output_dim, window_size, fit_intercept, normalize
):
    # NOTE: we use random data because we want to test dimensions and
    # correctness vs a second implementation
    X = onp.random.rand(n, input_dim * window_size)
    y = onp.random.rand(n, output_dim)

    ar = AR(
        input_dim,
        output_dim,
        window_size,
        fit_intercept=fit_intercept,
        normalize=normalize,
    )
    inv, Xy = ar._fit_accumulate(generator(X, y), alpha=1)
    beta = _ar_fit_unconstrained(inv, Xy)

    onp.testing.assert_array_almost_equal(
        beta,
        ols(
            X, y, input_dim, alpha=1, fit_intercept=fit_intercept, normalize=normalize,
        ),
        decimal=3,
    )


@pytest.mark.parametrize("n", [10, 1000])
@pytest.mark.parametrize("input_dim", [1, 10])
@pytest.mark.parametrize("output_dim", [1, 10])
@pytest.mark.parametrize("window_size", [1, 10])
@pytest.mark.parametrize("fit_intercept", [True, False])
def test_fit_constrained(n, input_dim, output_dim, window_size, fit_intercept):
    # NOTE: we use random data because we want to test dimensions and
    # correctness vs a second implementation
    print(n, input_dim, output_dim, window_size, fit_intercept)
    X = onp.random.rand(n, input_dim * window_size)
    y = onp.random.rand(n, output_dim)

    ar = AR(
        input_dim, output_dim, window_size, fit_intercept=fit_intercept, constrain=True
    )
    inv, Xy = ar._fit_accumulate(generator(X, y), alpha=1)
    beta = _ar_fit_unconstrained(inv, Xy)

    # First double check that unconstrained also works
    onp.testing.assert_array_almost_equal(
        beta,
        ols(X, y, input_dim, alpha=1, fit_intercept=fit_intercept, constrain=True),
        decimal=3,
    )

    # Next, check that each chunk of input_dim features have the same coefficient
    R, r = _ar_form_linear_constraints(
        input_dim, output_dim, window_size, fit_intercept=fit_intercept
    )
    result = _ar_fit_constrained(beta, inv, R, r)

    num_params = window_size + (1 if fit_intercept else 0)
    assert (
        np.sum([np.abs(x - x[0]) for x in result.reshape(num_params, input_dim, -1)])
        < 1e-4
    )

    # Finally, check that resulting vector is of the correct length and the
    # values are self-consistent
    ar = AR(
        input_dim, output_dim, window_size, fit_intercept=fit_intercept, constrain=True
    )
    ar.fit(generator(X, y))
    assert len(ar._params["phi"]) == num_params
    onp.testing.assert_array_almost_equal(
        ar._params["phi"],
        np.mean(result.reshape(num_params, input_dim, -1), axis=1),
        decimal=3,
    )


@pytest.mark.parametrize(
    "window_size,input_dim,output_dim,fit_intercept,expected_R,expected_r",
    [
        (
            3,
            2,
            1,
            False,
            np.array(
                [
                    [1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, -1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, -1.0],
                ]
            ),
            np.zeros(3),
        ),
        (
            3,
            2,
            1,
            True,
            np.array(
                [
                    [1.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, -1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, -1.0],
                ]
            ),
            np.zeros(4),
        ),
        (4, 1, 1, True, np.zeros((0, 5)), np.zeros((0))),
        (
            3,
            2,
            2,
            True,
            np.array(
                [
                    [1.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, -1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, -1.0],
                ]
            ),
            np.zeros((4, 2)),
        ),
    ],
)
def test_form_linear_constraints(
    window_size, input_dim, output_dim, fit_intercept, expected_R, expected_r
):
    R, r = _ar_form_linear_constraints(
        input_dim, output_dim, window_size, fit_intercept
    )

    r = r.squeeze()

    assert np.array_equal(expected_R, R)
    assert np.array_equal(expected_r, r)


@pytest.mark.parametrize(
    "num_observations,num_features", [(2, 1), (2, 10), (10, 1), (10, 10)]
)
def test_fit_normalize_xtx(num_observations, num_features):
    # TODO: Investigate numerical stability
    # NOTE: Number of observations must be greater than 1
    M = onp.random.randint(0, 500, (num_observations, num_features))
    mu = M.mean(axis=0)
    sigma = M.std(axis=0)

    M_stats = OnlineStatistics(M.shape[1])
    for m in M:
        M_stats.update(m)

    A = (M - mu) / sigma
    onp.testing.assert_array_almost_equal(
        np.nan_to_num(A.T @ A),
        _ar_fit_normalize_prod(M.T @ M, M_stats, M_stats),
        decimal=2,
    )


@pytest.mark.parametrize(
    "n,A_features,B_features",
    [(2, 1, 10), (2, 10, 1), (10, 1, 10), (10, 10, 1), (10, 100, 100)],
)
def test_fit_normalize_ab(n, A_features, B_features):
    # TODO: Investigate numerical stability
    # NOTE: Number of observations must be greater than 1
    A = onp.random.randint(0, 500, (n, A_features))
    A_mu = A.mean(axis=0)
    A_sigma = A.std(axis=0)

    A_stats = OnlineStatistics(A.shape[1])
    for a in A:
        A_stats.update(a)

    A_norm = (A - A_mu) / A_sigma

    B = onp.random.rand(n, B_features)
    B_mu = B.mean(axis=0)
    B_sigma = B.std(axis=0)

    B_stats = OnlineStatistics(B.shape[1])
    for b in B:
        B_stats.update(b)

    B_norm = (B - B_mu) / B_sigma

    onp.testing.assert_array_almost_equal(
        np.nan_to_num(A_norm.T @ B_norm),
        _ar_fit_normalize_prod(A.T @ B, A_stats, B_stats),
        decimal=2,
    )



from abc import ABC, abstractmethod

import jax.numpy as np


class Loss(ABC):
    @abstractmethod
    def __init__(self):
        raise NotImplementedError()

    @abstractmethod
    def compute(self, y_pred: np.ndarray, y_true: np.ndarray):
        raise NotImplementedError()



from abc import ABC, abstractmethod
from numbers import Real
from typing import Generator, TypeVar

import jax.numpy as np

# TODO: How to distinguish between a 1D time series value and a 1D time series?
# TODO: Enforce time series inputs
TimeSeries = TypeVar("TimeSeries", Real, np.ndarray)


class TimeSeriesMixin:
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._timeseries = True


class RegressorMixin:
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._regressor = True


class DifferentiableMixin:
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._differentiable = True


class OfflineMixin(ABC):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._offline = True

    @abstractmethod
    # TODO: Update documentation
    def fit(self, generator: Generator) -> "OfflineMixin":
        """
        Description: catch-all for training step (e.g., offline, pre-train,
            incremental).
        Args:
            data (TimeSeries):
        Returns:
            method (BaseMethod):
        """
        raise NotImplementedError()


class BaseLearner(ABC):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        """
        Description: Initialize BaseMethod
        Returns:
            method (BaseMethod):
        """
        super().__init__()

        # TODO: May want to check that certain keyword args are supplied
        for key in kwargs:
            if not hasattr(self, key):
                setattr(self, key, kwargs[key])

    @abstractmethod
    def predict(self, data: TimeSeries) -> TimeSeries:
        """
        Description: make a prediction about the next value based on current
            data.
        Args:
            data (TimeSeries):
        Returns:
            ts (TimeSeriesValue):
        """
        raise NotImplementedError()

    @abstractmethod
    # TODO: type hint y
    def update(self, data: TimeSeries, y) -> "BaseLearner":
        """
        Description: updates a model? What should this do?
        Args:
            data (TimeSeries):
        Returns:
            method (BaseMethod):
        """
        raise NotImplementedError()



from proust.learners._ar import AR
from proust.learners._black_box import BlackBox, DifferentiableBlackBox, blackbox
from proust.learners._boosted import Boosted, boost
from proust.learners._predict_constant import PredictConstant
from proust.learners._predict_last import PredictLast

__all__ = [
    "AR",
    "BlackBox",
    "Boosted",
    "boost",
    "DifferentiableBlackBox",
    "blackbox",
    "PredictConstant",
    "PredictLast",
]



from proust.utils.losses._cross_entropy import CrossEntropy
from proust.utils.losses._huber import Huber
from proust.utils.losses._mean_absolute_error import MeanAbsoluteError
from proust.utils.losses._mean_bias_error import MeanBiasError
from proust.utils.losses._mean_square_error import MeanSquareError

__all__ = [
    "CrossEntropy",
    "Huber",
    "MeanAbsoluteError",
    "MeanBiasError",
    "MeanSquareError",
]



from numbers import Real

import jax.numpy as np

from proust.learners.core import BaseLearner, TimeSeries, TimeSeriesMixin


class PredictLast(TimeSeriesMixin, BaseLearner):
    """
    Description: Predicts the value n steps ago in the time series, i.e.,
    x(t + 1) = x(t - n)
    """

    def __init__(self, offset: int = 1, pad_value: Real = 0):
        """
        Description: Initializes LastMethod
        Args:
            offset (int): number of steps to look back in the time series
                TODO: how to size the buffer?
            pad_value (Real): value used to pad missing predictions.
                TODO: what if each value is vector?
        Returns:
            method (LastMethod):
        """
        super().__init__()
        self._offset = offset
        self._pad_value = pad_value

        # TODO: we don't know how to size buffer yet
        self._buffer_ = []
        self._index = -1

    def predict(self, data: TimeSeries) -> TimeSeries:
        """
        TODO: Ignores the offset parameter
        Questions:
            - Does it make sense to take data in batch here except perhaps to
              advance the timeline by len number of steps?
        Args:
            data (TimeSeries):

        Returns:
            ts (TimeSeries):
        """
        if np.isscalar(data) or isinstance(data, np.ndarray):
            return data
        else:
            return np.asarray(data)

    def update(self, data: TimeSeries, y) -> "PredictLast":
        return self



import jax.numpy as np

from proust.utils.losses.core import Loss


class MeanAbsoluteError(Loss):
    def __init__(self):
        pass

    def compute(self, y_pred: np.ndarray, y_true: np.ndarray):
        """
        Description: mean absolute error (L1)
        Args:
            y_pred (np.ndarray): value predicted by method
            y_true (np.ndarray): ground truth value
        Returns:
            error (Real):
        """
        return np.sum(np.abs(y_pred - y_true)) / y_pred.shape[0]



from typing import List

from proust.learners.core import BaseLearner, TimeSeries


class Boosted(BaseLearner):
    def __init__(self, children: List[BaseLearner]):
        self._children = children

    def predict(self, data: TimeSeries) -> TimeSeries:
        pass

    def update(self, data: TimeSeries, y) -> "BaseLearner":
        pass


def boost(children: List[BaseLearner]) -> Boosted:
    return Boosted(children)



from numbers import Real
from typing import Tuple, Union

import jax.numpy as np


class OnlineStatistics:
    def __init__(self, input_dim: Union[Tuple[int], int]):
        self._input_dim = input_dim
        self._is_scalar = input_dim == 1
        self._mean = 0 if self._is_scalar else np.zeros(self._input_dim)
        self._variance = 0 if self._is_scalar else np.zeros(self._input_dim)
        self._sum = 0 if self._is_scalar else np.zeros(self._input_dim)
        self._observations = 0

    # TODO: Do we want to support batching?
    def update(self, observation: Union[Real, np.ndarray]) -> None:
        if not self._is_scalar:
            observation = np.asarray(observation)

        self._observations += 1
        prev_mean = self._mean
        self._mean = self._mean + (observation - self._mean) / self._observations
        self._variance = self._variance + (observation - prev_mean) * (
            observation - self._mean
        )
        self._sum += observation

    def mean(self) -> Union[Real, np.ndarray]:
        return self._mean

    def var(self) -> Union[Real, np.ndarray]:
        return self._variance / self._observations

    def std(self) -> Union[Real, np.ndarray]:
        return np.sqrt(self._variance / self._observations)

    def sum(self) -> Union[Real, np.ndarray]:
        return self._sum

    def observations(self) -> int:
        return self._observations

    def zscore(self, data: np.ndarray) -> Real:
        return (data - self.mean()) / self.std()



from numbers import Real

import jax.numpy as np

from proust.utils.losses.core import Loss


class Huber(Loss):
    def __init__(self, delta: Real = 2):
        self._delta = delta

    def compute(self, y_pred: np.ndarray, y_true: np.ndarray):
        """
        Description: Huber loss; square error when error is small, absolute
        error when large
        Args:
            y_pred (np.ndarray): value predicted by method
            y_true (np.ndarray): ground truth value
        Returns:
            error (Real):
        """
        diff = np.abs(y_pred - y_true)
        delta2 = self._delta * self._delta / 2

        return (
            np.sum(
                np.where(
                    diff <= self._delta, diff * diff / 2, self._delta * diff - delta2,
                )
            )
            / y_pred.shape[0]
        )



from numbers import Real

import jax.numpy as np

from proust.learners.core import BaseLearner, TimeSeries, TimeSeriesMixin


class PredictConstant(TimeSeriesMixin, BaseLearner):
    """
    Description: Predicts a constant value at each step, i.e., x(t + 1) = c
    """

    def __init__(self, constant: Real = 0):
        """
        Description: Initialize ConstantMethod
        Args:
            constant (Real): value to predict at each time step
        Returns:
            method (ConstantMethod):
        """
        super().__init__()
        self._constant = constant

    def predict(self, data: TimeSeries) -> TimeSeries:
        if np.isscalar(data):
            return self._constant
        else:
            return np.ones(np.asarray(data).shape) * self._constant

    def update(self, data: TimeSeries, y) -> "PredictConstant":
        return self



"""
Implements pseudorandomness in our program
"""

from random import randint

import jax.random

import proust

# NOTE: We use numpy because the jax RNG is deterministic


def set_key(key=None):
    """
    Description: Fix global random key to ensure reproducibility of results.

    Args:
        key (int): key that determines reproducible output
    """
    if key is None:
        key = randint(1, 9223372036854775807 + 1)
    assert type(key) == int
    proust.GLOBAL_RANDOM_KEY = jax.random.PRNGKey(key)


def generate_key():
    """
    Description: Generate random key.

    Returns:
        Random random key
    """
    key, subkey = jax.random.split(proust.GLOBAL_RANDOM_KEY)
    proust.GLOBAL_RANDOM_KEY = key
    return subkey


def get_global_key():
    """
    Description: Get current global random key.

    Returns:
        Current global random key
    """

    return proust.GLOBAL_RANDOM_KEY



from typing import Callable

import jax

from proust.learners.core import BaseLearner, DifferentiableMixin, TimeSeries
from proust.utils.losses import MeanSquareError
from proust.utils.losses.core import Loss
from proust.utils.optimizers import OGD
from proust.utils.optimizers.core import Optimizer


class BlackBox(BaseLearner):
    def __init__(self, predict: Callable):
        super().__init__()

        self._predict = predict

    def predict(self, data: TimeSeries) -> TimeSeries:
        # TODO: Assumes any featurizing is handled in self._predict
        return self._predict(data)

    def update(self, data: TimeSeries, y) -> "BlackBox":
        return self


class DifferentiableBlackBox(DifferentiableMixin, BlackBox):
    def __init__(
        self,
        predict: Callable,
        params: dict = None,
        optimizer: Optimizer = OGD,
        loss: Loss = None,
    ):
        super().__init__()

        self._predict = predict
        self._params = params
        self._optimizer = optimizer
        self._loss = MeanSquareError() if loss is None else loss

        try:
            self._predict = jax.jit(self._predict)
            self._objective = jax.jit(
                lambda _params, x, y: self._loss.compute(self._predict(_params, x), y)
            )
            self._grad = jax.jit(jax.grad(self._objective))
        except Exception as e:
            # TODO: Figure out actual jax exception
            raise ValueError(
                "Prediction function is not jitable via jax. Try"
                "`BlackBoxLearner` instead of"
                "`DifferentiableBlackBoxLearner`. {}".format(e)
            )

    def predict(self, data: TimeSeries) -> TimeSeries:
        # TODO: Assumes any featurizing is handled in self._predict
        return self._predict(self._params, data)

    def update(self, data: TimeSeries, y) -> "DifferentiableBlackBox":
        self._params = self._optimizer.update(
            self._params, self._grad(self._params, data, y)
        )
        return self


def blackbox(
    predict: Callable,
    params: dict = None,
    optimizer: Optimizer = OGD,
    loss: Loss = MeanSquareError,
) -> BlackBox:
    try:
        return DifferentiableBlackBox(predict, params, optimizer, loss)
    except Exception as e:
        e
        return BlackBox(predict)



from abc import ABC, abstractmethod


# TODO: Add type hints
class Optimizer(ABC):
    @abstractmethod
    def __init__(self):
        raise NotImplementedError()

    @abstractmethod
    def update(self, params, grad):
        raise NotImplementedError()



import jax.numpy as np
import pytest

from proust.learners._predict_last import PredictLast


@pytest.mark.parametrize(
    "x,y",
    [
        (0, 0),
        (4.3, 4.3),
        (np.array([0]), np.zeros(1)),
        (np.array([0, 1, 2, 3, 4]), np.asarray([0, 1, 2, 3, 4])),
    ],
)
def test_predict_last(x, y):
    predict_last = PredictLast()

    assert np.array_equal(predict_last.predict(x), y)



import jax.random

from proust.learners import blackbox, boost
from proust.utils.random import set_key

# Initialize global random key by seeding the jax random number generator
GLOBAL_RANDOM_KEY = jax.random.PRNGKey(0)
set_key()

__all__ = ["blackbox", "boost"]



import os

from setuptools import find_packages, setup

with open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md"),
    encoding="utf-8",
) as f:
    long_description = f.read()

setup(
    name="proust",
    version="0.1.2",
    author="Daniel Suo",
    author_email="danielsuo@gmail.com",
    description="Online time series analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/danielsuo/proust",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=["timeseries", "time", "series", "analysis"],
    python_requires=">=3.5",
    install_requires=[
        "numpy",
        "scipy",
        "cython",
        "jax",
        "jaxlib",
        "scikit-learn",
        "matplotlib",
    ],
    extras_require={
        "dev": [
            "flake8",
            "flake8-print",
            "flake8-bugbear",
            "mypy",
            "pytest",
            "pytest-xdist",
            "pytest-cov",
            "pylint",
            "black",
            "isort",
            "autoflake",
            "pre-commit",
            "pydocstring-coverage",
            "bumpversion",
        ]
    },
)



import jax.numpy as np

from proust.utils.optimizers.core import Optimizer


class OGD(Optimizer):
    def __init__(self, learning_rate=1.0, max_norm=True):
        # Create dictionary of named arguments
        super().__init__()

        self.T = 0
        self.max_norm = max_norm
        self.original_max_norm = self.max_norm
        self.learning_rate = learning_rate

    def update(self, params, grad):
        self.T += 1

        # TODO: Don't overload `learning_rate`
        # TODO: OGD / SGD grand unification theory
        # TODO: Root T
        learning_rate = self.learning_rate / np.sqrt(self.T)

        if self.max_norm:
            self.max_norm = np.maximum(
                self.max_norm,
                np.linalg.norm([np.linalg.norm(dw) for dw in grad.values()]),
            )
            learning_rate = self.learning_rate / self.max_norm

        new_params = {key: params[key] - learning_rate * grad[key] for key in params}

        return new_params
